@@ -19,36 +19,6 @@
 #include "goxel.h"
 #include "xxhash.h"
 
-/* History
-    the images undo history is stored in a linked list.  Every time we call
-    image_history_push, we add the current image snapshot in the list.
-
-    For example, if we did three operations, A, B, C, and now the image is
-    in the D state, the history list looks like this:
-
-    img->history                                        img
-        |                                                |
-        v                                                v
-    +--------+       +--------+       +--------+      +--------+
-    |        |       |        |       |        |      |        |
-    |   A    |------>|   B    |------>|   C    |----->|   D    |
-    |        |       |        |       |        |      |        |
-    +--------+       +--------+       +--------+      +--------+
-
-    After an undo, we get:
-
-    img->history                        img
-        |                                |
-        v                                v
-    +--------+       +--------+       +--------+     +--------+
-    |        |       |        |       |        |     |        |
-    |   A    |------>|   B    |------>|   C    |---->|   D    |
-    |        |       |        |       |        |     |        |
-    +--------+       +--------+       +--------+     +--------+
-
-
-*/
-
 
 static bool material_name_exists(void *user, const char *name)
 {
@@ -190,17 +160,82 @@ image_t *image_new(void)
     layer->id = img_get_new_id(img);
     layer->material = img->active_material;
     DL_APPEND(img->layers, layer);
-    DL_APPEND2(img->history, img, history_prev, history_next);
     img->active_layer = layer;
     // Prevent saving an empty image.
     img->saved_key = image_get_key(img);
+    image_history_push(img);
     return img;
 }
 
+static void image_restore(image_t *img, const image_t *snap)
+{
+    layer_t *layer, *snap_layer;
+    camera_t *camera, *snap_camera;
+    material_t *material, *snap_material;
+    // Note: this would be easier if we has some generic object array
+    // objects.
+
+    // First remove all the layer, cameras, and materials.
+    while ((layer = img->layers)) {
+        DL_DELETE(img->layers, layer);
+        layer_delete(layer);
+    }
+    while ((camera = img->cameras)) {
+        DL_DELETE(img->cameras, camera);
+        camera_delete(camera);
+    }
+    while ((material = img->materials)) {
+        DL_DELETE(img->materials, material);
+        material_delete(material);
+    }
+
+    // Set copy from the other.
+    img->layers = NULL;
+    img->active_layer = NULL;
+    DL_FOREACH(snap->layers, snap_layer) {
+        layer = layer_copy(snap_layer);
+        DL_APPEND(img->layers, layer);
+        if (snap_layer == snap->active_layer)
+            img->active_layer = layer;
+    }
+    assert(img->active_layer);
+
+    img->cameras = NULL;
+    img->active_camera = NULL;
+    DL_FOREACH(snap->cameras, snap_camera) {
+        camera = camera_copy(snap_camera);
+        DL_APPEND(img->cameras, camera);
+        if (snap_camera == snap->active_camera)
+            img->active_camera = camera;
+    }
+
+    img->materials = NULL;
+    img->active_material = NULL;
+    DL_FOREACH(snap->materials, snap_material) {
+        material = material_copy(snap_material);
+        DL_APPEND(img->materials, material);
+        if (snap_material == snap->active_material)
+            img->active_material = material;
+        DL_FOREACH(img->layers, layer) {
+            if (layer->material == snap_material)
+                layer->material = material;
+        }
+    }
+
+    // Copy other attributes.
+    mat4_copy(snap->box, img->box);
+    mat4_copy(snap->selection_box, img->selection_box);
+
+    volume_delete(img->selection_mask);
+    img->selection_mask = NULL;
+    if (snap->selection_mask)
+        img->selection_mask = volume_copy(snap->selection_mask);
+}
+
 /*
  * Generate a copy of the image that can be put into the history.
  */
-static image_t *image_snap(image_t *other)
+static image_t *image_snapshot(const image_t *other)
 {
     image_t *img;
     layer_t *layer, *other_layer;
@@ -208,10 +243,9 @@ static image_t *image_snap(image_t *other)
     material_t *material, *other_material;
 
     img = calloc(1, sizeof(*img));
-    *img = *other;
+    mat4_copy(other->box, img->box);
+    mat4_copy(other->selection_box, img->selection_box);
 
-    img->layers = NULL;
-    img->active_layer = NULL;
     DL_FOREACH(other->layers, other_layer) {
         layer = layer_copy(other_layer);
         DL_APPEND(img->layers, layer);
@@ -220,8 +254,6 @@ static image_t *image_snap(image_t *other)
     }
     assert(img->active_layer);
 
-    img->cameras = NULL;
-    img->active_camera = NULL;
     DL_FOREACH(other->cameras, other_camera) {
         camera = camera_copy(other_camera);
         DL_APPEND(img->cameras, camera);
@@ -229,8 +261,6 @@ static image_t *image_snap(image_t *other)
             img->active_camera = camera;
     }
 
-    img->materials = NULL;
-    img->active_material = NULL;
     DL_FOREACH(other->materials, other_material) {
         material = material_copy(other_material);
         DL_APPEND(img->materials, material);
@@ -242,14 +272,16 @@ static image_t *image_snap(image_t *other)
         }
     }
 
-    img->history = img->history_next = img->history_prev = NULL;
+    if (other->selection_mask)
+        img->selection_mask = volume_copy(other->selection_mask);
+
     return img;
 }
 
 
 void image_delete(image_t *img)
 {
-    image_t *hist, *snap, *snap_tmp;
+    image_t *snap, *snap_tmp;
     camera_t *cam;
     layer_t *layer;
     material_t *mat;
@@ -270,17 +302,12 @@ void image_delete(image_t *img)
         material_delete(mat);
     }
 
-    // Path is shared between images and snaps!
-    // XXX: find a better way.
-    if (img->history) {
-        free(img->path);
-        img->path = NULL;
-    }
+    free(img->path);
+    free(img->export_path);
 
-    hist = img->history;
-    DL_FOREACH_SAFE2(hist, snap, snap_tmp, history_next) {
-        if (snap == img) continue;
-        DL_DELETE2(hist, snap, history_prev, history_next);
+    DL_FOREACH_SAFE2 (img->history, snap, snap_tmp, history_next) {
+        DL_DELETE2(img->history, snap, history_prev, history_next);
+        assert(snap->ref <= 1);
         image_delete(snap);
     }
 
@@ -312,8 +339,8 @@ layer_t *image_add_shape_layer(image_t *img)
     layer->shape = &shape_sphere;
     vec4_copy(goxel.painter.color, layer->color);
     // If the selection is on use it, otherwise center it in the image.
-    if (!box_is_null(goxel.selection)) {
-        mat4_copy(goxel.selection, layer->mat);
+    if (!box_is_null(img->selection_box)) {
+        mat4_copy(img->selection_box, layer->mat);
     } else {
         vec3_copy(img->box[3], layer->mat[3]);
         mat4_iscale(layer->mat, 4, 4, 4);
@@ -420,7 +447,7 @@ void image_merge_visible_layers(image_t *img)
                 }
             }
             SWAP(layer->volume, last->volume);
-            volume_merge(layer->volume, last->volume, MODE_OVER, NULL);
+            volume_merge(layer->volume, last->volume, layer->mode, NULL);
             DL_DELETE(img->layers, last);
             layer_delete(last);
         }
@@ -429,6 +456,30 @@ void image_merge_visible_layers(image_t *img)
     if (last) img->active_layer = last;
 }
 
+void image_merge_layer_down(image_t *img, layer_t *layer)
+{
+    layer_t *target, *other;
+    assert(img);
+    assert(layer);
+
+    if (layer == img->layers) return; // First layer.
+    target = layer->prev;
+
+    image_unclone_layer(img, layer);
+    image_unclone_layer(img, target);
+
+    // Unclone all layers cloned from this one.
+    DL_FOREACH(goxel.image->layers, other) {
+        if (other->base_id == layer->id) {
+            other->base_id = 0;
+        }
+    }
+
+    volume_merge(target->volume, layer->volume, layer->mode, NULL);
+    DL_DELETE(img->layers, layer);
+    layer_delete(layer);
+    img->active_layer = target;
+}
 
 camera_t *image_add_camera(image_t *img, camera_t *cam)
 {
@@ -533,12 +584,14 @@ void image_set(image_t *img, image_t *other)
 }
 
 #if 0 // For debugging purpose.
-static void debug_print_history(image_t *img)
+void debug_print_history(image_t *img)
 {
     int i = 0;
+
     image_t *hist;
     DL_FOREACH2(img->history, hist, history_next) {
-        printf("%d%s  ", i++, hist == img ? "*" : " ");
+        assert(hist != img);
+        printf("%d%s  ", i++, hist == img->history_pos ? "*" : " ");
     }
     printf("\n");
 }
@@ -548,19 +601,27 @@ static void debug_print_history(image_t *img) {}
 
 void image_history_push(image_t *img)
 {
-    image_t *snap = image_snap(img);
-    image_t *hist;
+    image_t *snap;
 
-    // Discard previous undo.
-    while ((hist = img->history_next)) {
-        DL_DELETE2(img->history, hist, history_prev, history_next);
-        assert(hist != img->history_next);
-        image_delete(hist);
+    // Don't do anything if the image didn't actually changed.
+    if (img->history_pos) {
+        if (image_get_key(img) == image_get_key(img->history_pos)) {
+            return;
+        }
     }
 
-    DL_DELETE2(img->history, img,  history_prev, history_next);
+    // Discard previous undo.
+    if (img->history_pos) {
+        while (img->history_pos->history_next) {
+            snap = img->history_pos->history_next;
+            DL_DELETE2(img->history, snap, history_prev, history_next);
+            debug_print_history(img);
+        }
+    }
+
+    snap = image_snapshot(img);
     DL_APPEND2(img->history, snap, history_prev, history_next);
-    DL_APPEND2(img->history, img,  history_prev, history_next);
+    img->history_pos = snap;
     debug_print_history(img);
 }
 
@@ -588,32 +649,25 @@ void image_history_resize(image_t *img, int size)
     }
 }
 
-// XXX: not clear what this is doing.  We should try to remove it.
-// It swap the content of two images without touching their pointer or
-// history.
-static void swap(image_t *a, image_t *b)
-{
-    SWAP(*a, *b);
-    SWAP(a->history, b->history);
-    SWAP(a->history_next, b->history_next);
-    SWAP(a->history_prev, b->history_prev);
-}
-
 void image_undo(image_t *img)
 {
-    image_t *prev = img->history_prev;
-    if (img->history == img) {
+    camera_t camera = {};
+    image_t *prev;
+    if (img->history_pos == img->history) {
         LOG_D("No more undo");
+        debug_print_history(img);
         return;
     }
-    DL_DELETE2(img->history, img, history_prev, history_next);
-    DL_PREPEND_ELEM2(img->history, prev, img, history_prev, history_next);
-    swap(img, prev);
+    assert(img->history_pos);
+    if (img->active_camera) camera = *img->active_camera;
+    prev = img->history_pos->history_prev;
+    image_restore(img, prev);
+    img->history_pos = prev;
 
     // Don't move the camera for an undo.
-    if (img->active_camera && prev->active_camera &&
-            strcmp(img->active_camera->name, prev->active_camera->name) == 0) {
-        camera_set(img->active_camera, prev->active_camera);
+    if (img->active_camera &&
+            strcmp(img->active_camera->name, camera.name) == 0) {
+        camera_set(img->active_camera, &camera);
     }
 
     debug_print_history(img);
@@ -621,29 +675,31 @@ void image_undo(image_t *img)
 
 void image_redo(image_t *img)
 {
-    image_t *next = img->history_next;
-    if (!next) {
+    image_t *next;
+    if (!img->history_pos || !img->history_pos->history_next) {
         LOG_D("No more redo");
         return;
     }
-    DL_DELETE2(img->history, next, history_prev, history_next);
-    DL_PREPEND_ELEM2(img->history, img, next, history_prev, history_next);
-    swap(img, next);
+    next = img->history_pos->history_next;
+    image_restore(img, next);
+    img->history_pos = next;
     debug_print_history(img);
 }
 
 static void image_clear_layer(void)
 {
     painter_t painter;
-    layer_t *layer = goxel.image->active_layer;
-    if (box_is_null(goxel.selection) && volume_is_empty(goxel.mask)) {
+    image_t *img = goxel.image;
+    layer_t *layer = img->active_layer;
+    if (    box_is_null(img->selection_box) &&
+            volume_is_empty(img->selection_mask)) {
         volume_clear(layer->volume);
         return;
     }
 
     // Use the mask in priority if it exists.
-    if (!volume_is_empty(goxel.mask)) {
-        volume_merge(layer->volume, goxel.mask, MODE_SUB, NULL);
+    if (!volume_is_empty(img->selection_mask)) {
+        volume_merge(layer->volume, img->selection_mask, MODE_SUB, NULL);
         return;
     }
 
@@ -652,7 +708,7 @@ static void image_clear_layer(void)
         .mode = MODE_SUB,
         .color = {255, 255, 255, 255},
     };
-    volume_op(layer->volume, &painter, goxel.selection);
+    volume_op(layer->volume, &painter, img->selection_box);
 }
 
 bool image_layer_can_edit(const image_t *img, const layer_t *layer)
@@ -683,6 +739,9 @@ uint32_t image_get_key(const image_t *img)
         k = material_get_hash(material);
         key = XXH32(&k, sizeof(k), key);
     }
+    key = XXH32(img->selection_box, sizeof(img->selection_box), key);
+    k = volume_get_key(img->selection_mask);
+    key = XXH32(&k, sizeof(k), key);
     return key;
 }
 
@@ -728,8 +787,7 @@ static void image_image_layer_to_volume(image_t *img, layer_t *layer)
     free(data);
 }
 
-ACTION_REGISTER(layer_clear,
-    .help = "Clear the current layer",
+ACTION_REGISTER(ACTION_layer_clear,
     .cfunc = image_clear_layer,
     .icon = ICON_DELETE,
     .flags = ACTION_TOUCH_IMAGE,
@@ -741,8 +799,7 @@ static void a_image_add_layer(void)
     image_add_layer(goxel.image, NULL);
 }
 
-ACTION_REGISTER(img_new_layer,
-    .help = "Add a new layer to the image",
+ACTION_REGISTER(ACTION_img_new_layer,
     .cfunc = a_image_add_layer,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_ADD,
@@ -753,8 +810,7 @@ static void a_image_delete_layer(void)
     image_delete_layer(goxel.image, goxel.image->active_layer);
 }
 
-ACTION_REGISTER(img_del_layer,
-    .help = "Delete the active layer",
+ACTION_REGISTER(ACTION_img_del_layer,
     .cfunc = a_image_delete_layer,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_REMOVE,
@@ -771,15 +827,13 @@ static void a_image_move_layer_down(void)
 }
 
 
-ACTION_REGISTER(img_move_layer_up,
-    .help = "Move the active layer up",
+ACTION_REGISTER(ACTION_img_move_layer_up,
     .cfunc = a_image_move_layer_up,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_ARROW_UPWARD,
 )
 
-ACTION_REGISTER(img_move_layer_down,
-    .help = "Move the active layer down",
+ACTION_REGISTER(ACTION_img_move_layer_down,
     .cfunc = a_image_move_layer_down,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_ARROW_DOWNWARD,
@@ -790,8 +844,7 @@ static void a_image_duplicate_layer(void)
     image_duplicate_layer(goxel.image, goxel.image->active_layer);
 }
 
-ACTION_REGISTER(img_duplicate_layer,
-    .help = "Duplicate the active layer",
+ACTION_REGISTER(ACTION_img_duplicate_layer,
     .cfunc = a_image_duplicate_layer,
     .flags = ACTION_TOUCH_IMAGE,
 )
@@ -801,8 +854,8 @@ static void a_image_clone_layer(void)
     image_clone_layer(goxel.image, goxel.image->active_layer);
 }
 
-ACTION_REGISTER(img_clone_layer,
-    .help = "Clone the active layer",
+ACTION_REGISTER(ACTION_img_clone_layer,
+    .help = N_("Creates a linked copy of the selected layer"),
     .cfunc = a_image_clone_layer,
     .flags = ACTION_TOUCH_IMAGE,
 )
@@ -812,8 +865,8 @@ static void a_image_unclone_layer(void)
     image_unclone_layer(goxel.image, goxel.image->active_layer);
 }
 
-ACTION_REGISTER(img_unclone_layer,
-    .help = "Unclone the active layer",
+ACTION_REGISTER(ACTION_img_unclone_layer,
+    .help = N_("Unlinks the selected layer from its parent"),
     .cfunc = a_image_unclone_layer,
     .flags = ACTION_TOUCH_IMAGE,
 )
@@ -825,8 +878,7 @@ static void a_img_select_parent_layer(void)
 }
 
 
-ACTION_REGISTER(img_select_parent_layer,
-    .help = "Select the parent of a layer",
+ACTION_REGISTER(ACTION_img_select_parent_layer,
     .cfunc = a_img_select_parent_layer,
     .flags = ACTION_TOUCH_IMAGE,
 )
@@ -836,19 +888,27 @@ static void a_img_merge_visible_layers(void)
     image_merge_visible_layers(goxel.image);
 }
 
-ACTION_REGISTER(img_merge_visible_layers,
-    .help = "Merge all the visible layers",
+ACTION_REGISTER(ACTION_img_merge_visible_layers,
     .cfunc = a_img_merge_visible_layers,
     .flags = ACTION_TOUCH_IMAGE,
 )
 
+static void a_img_merge_layer_down(void)
+{
+    image_merge_layer_down(goxel.image, goxel.image->active_layer);
+}
+
+ACTION_REGISTER(ACTION_img_merge_layer_down,
+    .cfunc = a_img_merge_layer_down,
+    .flags = ACTION_TOUCH_IMAGE,
+)
+
 static void a_img_new_camera(void)
 {
     image_add_camera(goxel.image, NULL);
 }
 
-ACTION_REGISTER(img_new_camera,
-    .help = "Add a new camera to the image",
+ACTION_REGISTER(ACTION_img_new_camera,
     .cfunc = a_img_new_camera,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_ADD,
@@ -859,8 +919,7 @@ static void a_img_del_camera(void)
     image_delete_camera(goxel.image, goxel.image->active_camera);
 }
 
-ACTION_REGISTER(img_del_camera,
-    .help = "Delete the active camera",
+ACTION_REGISTER(ACTION_img_del_camera,
     .cfunc = a_img_del_camera,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_REMOVE,
@@ -876,15 +935,13 @@ static void a_img_move_camera_down(void)
     image_move_camera(goxel.image, goxel.image->active_camera, -1);
 }
 
-ACTION_REGISTER(img_move_camera_up,
-    .help = "Move the active camera up",
+ACTION_REGISTER(ACTION_img_move_camera_up,
     .cfunc = a_img_move_camera_up,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_ARROW_UPWARD,
 )
 
-ACTION_REGISTER(img_move_camera_down,
-    .help = "Move the active camera down",
+ACTION_REGISTER(ACTION_img_move_camera_down,
     .cfunc = a_img_move_camera_down,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_ARROW_DOWNWARD,
@@ -895,8 +952,8 @@ static void a_img_image_layer_to_volume(void)
     image_image_layer_to_volume(goxel.image, goxel.image->active_layer);
 }
 
-ACTION_REGISTER(img_image_layer_to_volume,
-    .help = "Turn an image layer into a volume",
+ACTION_REGISTER(ACTION_img_image_layer_to_volume,
+    .help = N_("Converts a layer into a volume"),
     .cfunc = a_img_image_layer_to_volume,
     .flags = ACTION_TOUCH_IMAGE,
 )
@@ -906,8 +963,8 @@ static void a_img_new_shape_layer(void)
     image_add_shape_layer(goxel.image);
 }
 
-ACTION_REGISTER(img_new_shape_layer,
-    .help = "Add a new shape layer to the image",
+ACTION_REGISTER(ACTION_img_new_shape_layer,
+    .help = N_("Creates a shape layer"),
     .cfunc = a_img_new_shape_layer,
     .flags = ACTION_TOUCH_IMAGE,
 )
@@ -917,8 +974,7 @@ static void a_img_new_material(void)
     image_add_material(goxel.image, NULL);
 }
 
-ACTION_REGISTER(img_new_material,
-    .help = "Add a new material to the image",
+ACTION_REGISTER(ACTION_img_new_material,
     .cfunc = a_img_new_material,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_ADD,
@@ -929,15 +985,14 @@ static void a_img_del_material(void)
     image_delete_material(goxel.image, goxel.image->active_material);
 }
 
-ACTION_REGISTER(img_del_material,
-    .help = "Delete a material",
+ACTION_REGISTER(ACTION_img_del_material,
     .cfunc = a_img_del_material,
     .flags = ACTION_TOUCH_IMAGE,
     .icon = ICON_REMOVE,
 )
 
-ACTION_REGISTER(img_auto_resize,
-    .help = "Auto resize the image to fit the layers",
+ACTION_REGISTER(ACTION_img_auto_resize,
+    .help = N_("Resizes the image to fit the layers"),
     .cfunc = a_image_auto_resize,
     .flags = ACTION_TOUCH_IMAGE,
 )
